// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#include <TestSIMD_MathOps.hpp>
#include <TestSIMD_MaskOps.hpp>
#include <TestSIMD_Conversions.hpp>
#include <TestSIMD_ShiftOps.hpp>
#include <TestSIMD_Condition.hpp>
#include <TestSIMD_GeneratorCtors.hpp>
#include <TestSIMD_Reductions.hpp>
#include <TestSIMD_Construction.hpp>
#include <TestSIMD_LoadStore.hpp>
