// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#ifndef KOKKOS_DECLARE_HPX_HPP
#define KOKKOS_DECLARE_HPX_HPP

#if defined(KOKKOS_ENABLE_HPX)
#include <HPX/Kokkos_HPX.hpp>
#include <HPX/Kokkos_HPX_MDRangePolicy.hpp>
#endif

#endif
