// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#ifndef KOKKOS_TEST_RANGEPOLICY_HPP
#define KOKKOS_TEST_RANGEPOLICY_HPP
#include <TestRange.hpp>
#endif
