// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#define KOKKOS_MATHEMATICAL_FUNCTIONS_SKIP_1
#define KOKKOS_MATHEMATICAL_FUNCTIONS_SKIP_3
#include "TestMathematicalFunctions.hpp"
#undef KOKKOS_MATHEMATICAL_FUNCTIONS_SKIP_1
#undef KOKKOS_MATHEMATICAL_FUNCTIONS_SKIP_3
