// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#ifndef KOKKOS_THREADS_FWD_HPP_
#define KOKKOS_THREADS_FWD_HPP_

#if defined(KOKKOS_ENABLE_THREADS)
namespace Kokkos {
class Threads;  ///< Execution space with C++11 threads back-end.
}  // namespace Kokkos
#endif
#endif
