// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#ifndef KOKKOSP_DEVICE_INFO_HPP
#define KOKKOSP_DEVICE_INFO_HPP

#include <cstdint>
#include <impl/Kokkos_Profiling_C_Interface.h>
namespace Kokkos {
namespace Profiling {
using KokkosPDeviceInfo = Kokkos_Profiling_KokkosPDeviceInfo;
}  // namespace Profiling
}  // namespace Kokkos

#endif
