// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#ifndef KOKKOS_TEST_SYCL_SHARED_USM_SPACE_HPP
#define KOKKOS_TEST_SYCL_SHARED_USM_SPACE_HPP

#include <gtest/gtest.h>

#define TEST_CATEGORY sycl_shared_usm
#define TEST_CATEGORY_DEATH sycl_shared_usm_DeathTest
#define TEST_EXECSPACE Kokkos::SYCLSharedUSMSpace

#endif
