// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#ifndef KOKKOS_ZEROMEMSET_FWD_HPP
#define KOKKOS_ZEROMEMSET_FWD_HPP

namespace Kokkos {
namespace Impl {

template <typename ExecutionSpace>
struct ZeroMemset;

}  // namespace Impl
}  // namespace Kokkos

#endif  // #ifndef KOKKOS_ZEROMEMSET_FWD_HPP
