#ifdef DESUL_IMPL_ATOMIC_CUDA_PTX_ISGLOBAL
#include "cuda_cc7_asm_atomic_fetch_op.inc_isglobal"
#endif

#ifdef DESUL_IMPL_ATOMIC_CUDA_PTX_PREDICATE
#include "cuda_cc7_asm_atomic_fetch_op.inc_predicate"
#endif

