// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#ifndef KOKKOS_OPENMP_FWD_HPP_
#define KOKKOS_OPENMP_FWD_HPP_

#if defined(KOKKOS_ENABLE_OPENMP)
namespace Kokkos {
class OpenMP;  ///< OpenMP execution space.
}  // namespace Kokkos
#endif
#endif
