// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#ifndef KOKKOS_HPX_FWD_HPP_
#define KOKKOS_HPX_FWD_HPP_

#if defined(KOKKOS_ENABLE_HPX)
namespace Kokkos {
namespace Experimental {
class HPX;  ///< Execution space with HPX back-end.
}  // namespace Experimental
}  // namespace Kokkos
#endif
#endif
