// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

/**
 * Header file to include all of Kokkos Tooling support
 */

#ifndef KOKKOS_IMPL_KOKKOS_TOOLS_HPP
#define KOKKOS_IMPL_KOKKOS_TOOLS_HPP

#include <impl/Kokkos_Profiling.hpp>

#endif
