// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#ifndef KOKKOS_SERIAL_FWD_HPP_
#define KOKKOS_SERIAL_FWD_HPP_

#if defined(KOKKOS_ENABLE_SERIAL)
namespace Kokkos {
class Serial;  ///< Execution space main process on CPU.
}  // namespace Kokkos
#endif
#endif
