// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#ifndef KOKKOS_TEST_CUDAUVM_HPP
#define KOKKOS_TEST_CUDAUVM_HPP

#include <gtest/gtest.h>

#define TEST_CATEGORY cuda_uvm
#define TEST_CATEGORY_DEATH cuda_uvm_DeathTest
#define TEST_EXECSPACE Kokkos::CudaUVMSpace

#endif
