// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#ifndef KOKKOS_TEST_OTHER_HPP
#define KOKKOS_TEST_OTHER_HPP
#include <TestMemoryPool.hpp>
#include <TestCXX11.hpp>

#ifdef KOKKOS_ENABLE_IMPL_VIEW_LEGACY
#include <TestViewCtorPropEmbeddedDim.hpp>
#endif
#endif
