/*
Copyright (c) 2019, Lawrence Livermore National Security, LLC
and DESUL project contributors. See the COPYRIGHT file for details.
Source: https://github.com/desul/desul

SPDX-License-Identifier: (BSD-3-Clause)
*/

#ifndef DESUL_ATOMICS_HPP_
#define DESUL_ATOMICS_HPP_

#include "desul/atomics/Atomic_Ref.hpp"
#include "desul/atomics/Compare_Exchange.hpp"
#include "desul/atomics/Generic.hpp"
#include "desul/atomics/Lock_Array.hpp"
#include "desul/atomics/Macros.hpp"

#endif
