// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
// SPDX-FileCopyrightText: Copyright Contributors to the Kokkos project

#ifndef KOKKOS_TEST_REDUCTIONS_HPP
#define KOKKOS_TEST_REDUCTIONS_HPP
#include <Kokkos_Macros.hpp>
#include <TestReduce.hpp>
#include <TestCXX11Deduction.hpp>
#endif
